import * as v from 'valibot';
import {ApplicationState, FALLBACK_CHARSET, FALLBACK_LANGUAGE, FALLBACK_MEDIA_TYPE, Resource} from '../common';
import http from 'http';
import {createServer, CreateServerParams} from './http/server';
import https from 'https';
import {BackendState} from './common';
import {BaseDataSource} from '../common/data-source';
import {DataSource} from './data-source';

export interface BackendResource<
	DataSourceType extends BaseDataSource = DataSource,
	ResourceSchema extends v.BaseSchema = v.BaseSchema,
	ResourceName extends string = string,
	ResourceRouteName extends string = string,
	IdAttr extends string = string,
	IdSchema extends v.BaseSchema = v.BaseSchema
> extends Resource<ResourceSchema, ResourceName, ResourceRouteName, IdAttr, IdSchema> {
	newId(dataSource: DataSourceType): string | number | unknown;
	dataSource: DataSourceType;
}

export interface BackendBuilder<T extends BaseDataSource = BaseDataSource> {
	showTotalItemCountOnGetCollection(b?: boolean): this;
	showTotalItemCountOnCreateItem(b?: boolean): this;
	checksSerializersOnDelete(b?: boolean): this;
	throwsErrorOnDeletingNotFound(b?: boolean): this;
	createHttpServer(serverParams?: CreateServerParams): http.Server | https.Server;
	dataSource?: (resource: Resource) => T;
}

export interface CreateBackendParams {
	app: ApplicationState;
	dataSource: (resource: Resource) => BaseDataSource;
}

export const createBackend = (params: CreateBackendParams) => {
	const backendState: BackendState = {
		app: params.app,
		dataSource: params.dataSource,
		cn: {
			language: FALLBACK_LANGUAGE,
			charset: FALLBACK_CHARSET,
			mediaType: FALLBACK_MEDIA_TYPE,
		},
		showTotalItemCountOnGetCollection: false,
		showTotalItemCountOnCreateItem: false,
		throws404OnDeletingNotFound: false,
		checksSerializersOnDelete: false,
	};

	return {
		showTotalItemCountOnGetCollection(b = true) {
			backendState.showTotalItemCountOnGetCollection = b;
			return this;
		},
		showTotalItemCountOnCreateItem(b = true) {
			backendState.showTotalItemCountOnCreateItem = b;
			return this;
		},
		throwsErrorOnDeletingNotFound(b = true) {
			backendState.throws404OnDeletingNotFound = b;
			return this;
		},
		checksSerializersOnDelete(b = true) {
			backendState.checksSerializersOnDelete = b;
			return this;
		},
		createHttpServer(serverParams = {} as CreateServerParams) {
			return createServer(backendState, serverParams);
		},
	} satisfies BackendBuilder;
};
