import {readFile, writeFile} from 'fs/promises';
import {join} from 'path';
import {DataSource as DataSourceInterface} from '../data-source';
import {Resource} from '../..';

export class DataSource<T extends Record<string, string>> implements DataSourceInterface<T> {
	private readonly path: string;

	data: T[] = [];

	constructor(private readonly resource: Resource,  baseDir = '') {
		this.path = join(baseDir, `${this.resource.state.routeName}.jsonl`);
	}

	async initialize() {
		try {
			const fileContents = await readFile(this.path, 'utf-8');
			const lines = fileContents.split('\n');
			this.data = lines.filter((l) => l.trim().length > 0).map((l) => JSON.parse(l));
		} catch (err) {
			await writeFile(this.path, '');
		}
	}

	async getTotalCount() {
		return this.data.length;
	}

	async getMultiple() {
		return [...this.data];
	}

	async getById(idSerialized: string) {
		const id = this.resource.state.idConfig.deserialize(idSerialized);
		const foundData = this.data.find((s) => s[this.resource.state.idAttr as string] === id);

		if (foundData) {
			return {
				...foundData
			};
		}

		return null;
	}

	async create(data: T) {
		const newData = {
			...data
		} as Record<string, unknown>;

		if (this.resource.state.idAttr in newData) {
			newData[this.resource.state.idAttr] = this.resource.state.idConfig.deserialize(newData[this.resource.state.idAttr] as string);
		}

		const newCollection = [
			...this.data,
			newData
		];

		await writeFile(this.path, newCollection.map((d) => JSON.stringify(d)).join('\n'));

		return data as T;
	}

	async delete(idSerialized: string) {
		const oldDataLength = this.data.length;

		const id = this.resource.state.idConfig.deserialize(idSerialized);
		const newData = this.data.filter((s) => !(s[this.resource.state.idAttr] === id));

		await writeFile(this.path, newData.map((d) => JSON.stringify(d)).join('\n'));

		return oldDataLength !== newData.length;
	}

	async emplace(idSerialized: string, dataWithId: T) {
		const existing = await this.getById(idSerialized);
		const id = this.resource.state.idConfig.deserialize(idSerialized);
		const { [this.resource.state.idAttr]: idFromResource, ...data } = dataWithId;
		const dataToEmplace = {
			...data,
			[this.resource.state.idAttr]: id,
		} as T;

		if (existing) {
			const newData = this.data.map((d) => {
				if (d[this.resource.state.idAttr] === id) {
					return dataToEmplace;
				}

				return d;
			});

			await writeFile(this.path, newData.map((d) => JSON.stringify(d)).join('\n'));

			return [dataToEmplace, false] as [T, boolean];
		}

		const newData = await this.create(dataToEmplace);
		return [newData, true] as [T, boolean];
	}

	async patch(idSerialized: string, data: Partial<T>) {
		const existing = await this.getById(idSerialized);
		if (!existing) {
			return null;
		}

		const newItem = {
			...existing,
			...data,
		}

		const id = this.resource.state.idConfig.deserialize(idSerialized);
		const newData = this.data.map((d) => {
			if (d[this.resource.state.idAttr as string] === id) {
				return newItem;
			}

			return d;
		});

		await writeFile(this.path, newData.map((d) => JSON.stringify(d)).join('\n'));
		return newItem as T;
	}
}
